/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*===========================================================================
 * 
 * Project : freyja
 * Author  : Terry 'Mongoose' Hendrix II
 * Website : http://www.icculus.org/~mongoose/
 * Email   : mongoose@icculus.org
 * Object  : SnippetsControl
 * License : No use w/o permission (C) 2007 Mongoose
 * Comments: This is the snippets interface class.
 *
 *
 *           This file was generated using Mongoose's C++ 
 *           template generator script.  <mongooseichiban@gmail.com>
 * 
 *-- History ------------------------------------------------- 
 *
 * 2007.06.15:
 * Mongoose - Created
 ==========================================================================*/

#include "SnippetsControl.h"


////////////////////////////////////////////////////////////
// Constructors
////////////////////////////////////////////////////////////

SnippetsControl::SnippetsControl()
{
}


SnippetsControl::~SnippetsControl()
{
}


////////////////////////////////////////////////////////////
// Public Accessors
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// Public Mutators
////////////////////////////////////////////////////////////

void SnippetsControl::EvSerializeMesh()
{
#if TINYXML_FOUND
	Mesh *m = Mesh::GetMesh( GetSelectedMesh() );

	if (!m)
		return;

	mstl::String path = freyja_rc_map_string("/");

	char *filename =
	mgtk_filechooser_blocking("freyja - Save Selected Mesh...", 
							  path.c_str(), 1,
							  "Mesh XML (*.xml)", "*.xml");

	if (!filename)
		return;

	TiXmlDocument doc;
	TiXmlDeclaration *decl = new TiXmlDeclaration("1.0", "", "");
	doc.LinkEndChild(decl);

	TiXmlElement *container = new TiXmlElement("freyja");
	container->SetAttribute("version", VERSION);
	container->SetAttribute("build-date", __DATE__);
	container->SetAttribute("file-version", 1);
	doc.LinkEndChild(container);

	for (uint32 i = 0, n = freyjaGetMaterialCount(); i < n; ++i)
	{
		Material *mat = freyjaGetMaterialClass(i);

		if (mat && i) // skip 'boring material'
			mat->Serialize(container);
	}

	m->Repack();

	if (filename &&  m->Serialize(container) && doc.SaveFile(filename))
	{
		mMeshXML.AddFilename(filename);
		Print("Mesh '%s' Saved", filename);
	}
	else if (filename)
	{
		Print("Mesh '%s' failed to save.", filename);
	}

	mgtk_filechooser_blocking_free(filename);
#endif
}


bool SnippetsControl::UnserializeMesh(const char *filename)
{
#if TINYXML_FOUND
	if (!filename)
		return false;

	TiXmlDocument doc(filename);

	if ( !doc.LoadFile() )
	{
		printf("XML ERROR: %s, Line %i, Col %i\n", 
			   doc.ErrorDesc(), doc.ErrorRow(), doc.ErrorCol() );
		return false;
	}

	TiXmlElement *root = doc.RootElement(); 

	if (!root) 
	{
		printf("Couldn't find document root!\n");
		return false;
	}

	TiXmlElement *child = root->FirstChildElement();
	for( ; child; child = child->NextSiblingElement() )
	{
		String s = child->Value();

		if (s == "mesh" || s == "Mesh")
		{
			Mesh *m = new Mesh();
			m->AddToPool();

		   	if ( m->Unserialize( child ) )
			{
				Print("Mesh '%s' Loaded", filename);
			}
			else
			{
				Print("Mesh '%s' failed to load.  <%s>", filename, 
					  child ? child->Value() : "(null)");
			}
		}
		else if (s == "material")
		{
			index_t idx = freyjaMaterialCreate();
			Material *mat = freyjaGetMaterialClass(idx);
			mat->Unserialize(child);
		}
	}

	mMeshXML.AddFilename(filename);

	return true;
#else
	return false;
#endif
}

void SnippetsControl::EvUnserializeMesh()
{
#if TINYXML_FOUND
	static mstl::String path = freyja_rc_map_string("/"); // cheesy
	char *filename =
	mgtk_filechooser_blocking("freyja - Open Selected Mesh...", 
							  path.c_str(), 0,
							  "Mesh XML (*.xml)", "*.xml");

	if ( UnserializeMesh(filename) )
	{
		path = filename;

		int off = path.find_last_of('/');

		if (off > 0)
		{
			path[off] = 0;
		}

		//Print("-- %s", path.c_str() );
		Print("Loaded snippet '%s'", filename);
	}

	mgtk_filechooser_blocking_free(filename);
#endif
}


bool SnippetsControl::SerializeBones(const char *filename)
{
#if TINYXML_FOUND
	TiXmlDocument doc;
	TiXmlDeclaration *decl = new TiXmlDeclaration("1.0", "", "");
	doc.LinkEndChild(decl);

	TiXmlElement *container = new TiXmlElement("freyja");
	container->SetAttribute("version", VERSION);
	container->SetAttribute("build-date", __DATE__);
	container->SetAttribute("file-version", 1);
	doc.LinkEndChild(container);

	Bone::SerializePool(container);

	mSkeletonXML.AddFilename(filename);
	return doc.SaveFile(filename);;
#else
	return false;
#endif
}


bool SnippetsControl::UnserializeBones(const char *filename)
{
#if TINYXML_FOUND
	TiXmlDocument doc(filename);

	if ( !doc.LoadFile() )
	{
		printf("XML ERROR: %s, Line %i, Col %i\n", 
			   doc.ErrorDesc(), doc.ErrorRow(), doc.ErrorCol() );
		return false;
	}

	TiXmlElement *root = doc.RootElement(); 

	if (!root) 
	{
		printf("Couldn't find document root!\n");
		return false;
	}

	//TiXmlElement *container = root->FirstChildElement();

	if ( Bone::UnserializePool(root) )
	{
		if (freyjaGetCurrentSkeleton() == INDEX_INVALID &&
		    freyjaGetSkeletonCount() > 0)
		{
			freyjaCurrentSkeleton(0);
		}

		UpdateSkeletalUI();	
		mSkeletonXML.AddFilename(filename);
		return true;
	}
	
	return false;
#else
	return false;
#endif
}


void SnippetsControl::EvUnserializeBones()
{
	static mstl::String path = freyja_rc_map_string("/"); // cheesy
	char *filename =
	mgtk_filechooser_blocking("freyja - Load Bones...", 
							  path.c_str(), 0,
							  "Bones XML (*.xml)", "*.xml");

	if (filename && UnserializeBones(filename) )
	{
		path = filename;

		int off = path.find_last_of('/');

		if (off > 0)
		{
			path[off] = 0;
		}

		//Print("-- %s", path.c_str() );
		Print("Loaded snippet '%s'", filename);
	}
	else if (filename)
	{
		Print("Bones '%s' failed to load.", filename);
	}
	
	mgtk_filechooser_blocking_free(filename);
}


void SnippetsControl::EvSerializeBones()
{
	mstl::String path = freyja_rc_map_string("/");
	char *filename =
	mgtk_filechooser_blocking("freyja - Save Bones...", 
							  path.c_str(), 1,
							  "Bones XML (*.xml)", "*.xml");

	if (filename && SerializeBones(filename) )
	{
		Print("Bones '%s' Saved", filename);
	}
	else if (filename)
	{
		Print("Bones '%s' failed to save.", filename);
	}
	
	mgtk_filechooser_blocking_free(filename);
}


////////////////////////////////////////////////////////////
// Private Accessors
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// Private Mutators
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// Unit Test code
////////////////////////////////////////////////////////////

#ifdef UNIT_TEST_SNIPPETSCONTROL
int runSnippetsControlUnitTest(int argc, char *argv[])
{
	SnippetsControl test;

	return 0;
}


int main(int argc, char *argv[])
{
	printf("[SnippetsControl class test]\n");

	return runSnippetsControlUnitTest(argc, argv);
}
#endif
